@@ -63,6 +63,9 @@ class DialogueTextTags(object):
         # Does this statement have a done tag?
         self.has_done = False
 
+        # Does this statement have a fast tag?
+        self.fast = False
+
         i = iter(TAG_RE.split(s))
 
         while True:
@@ -96,6 +99,7 @@ class DialogueTextTags(object):
                     self.pause_end = [ ]
                     self.pause_delay = [ ]
                     self.no_wait = False
+                    self.fast = True
 
                 elif tag == "done":
                     self.has_done = True
@@ -371,8 +375,9 @@ def show_display_say(who, what, who_args={}, what_args={}, window_args={},
 class SlowDone(object):
     delay = None
     ctc_kwargs = { }
+    last_pause = True
 
-    def __init__(self, ctc, ctc_position, callback, interact, type, cb_args, delay, ctc_kwargs): # @ReservedAssignment
+    def __init__(self, ctc, ctc_position, callback, interact, type, cb_args, delay, ctc_kwargs, last_pause): # @ReservedAssignment
         self.ctc = ctc
         self.ctc_position = ctc_position
         self.callback = callback
@@ -381,6 +386,7 @@ class SlowDone(object):
         self.cb_args = cb_args
         self.delay = delay
         self.ctc_kwargs = ctc_kwargs
+        self.last_pause = last_pause
 
     def __call__(self):
 
@@ -401,7 +407,7 @@ class SlowDone(object):
                 renpy.exports.restart_interaction()
 
         if self.delay is not None:
-            renpy.ui.pausebehavior(self.delay, True, voice=True)
+            renpy.ui.pausebehavior(self.delay, True, voice=self.last_pause)
             renpy.exports.restart_interaction()
 
         for c in self.callback:
@@ -516,12 +522,20 @@ def display_say(
 
     exception = None
 
+    if dtt.fast:
+        for i in renpy.config.say_sustain_callbacks:
+            i()
+
     try:
 
         for i, (start, end, delay) in enumerate(zip(pause_start, pause_end, pause_delay)):
 
+            # True if the is the last pause in a line of dialogue.
             last_pause = (i == len(pause_start) - 1)
 
+            if dtt.no_wait:
+                last_pause = False
+
             # If we're going to do an interaction, then saybehavior needs
             # to be here.
             if advance:
@@ -560,15 +574,21 @@ def display_say(
                 what_ctc = what_ctc._duplicate(None)
                 what_ctc._unique()
 
+            if ctc is not what_ctc:
+                if (ctc is not None) and ctc._duplicatable:
+                    ctc = ctc._duplicate(None)
+                    ctc._unique()
+
             if delay == 0:
                 what_ctc = None
+                ctc = None
 
             # Run the show callback.
             for c in callback:
                 c("show", interact=interact, type=type, **cb_args)
 
             # Create the callback that is called when the slow text is done.
-            slow_done = SlowDone(what_ctc, ctc_position, callback, interact, type, cb_args, delay, ctc_kwargs)
+            slow_done = SlowDone(what_ctc, ctc_position, callback, interact, type, cb_args, delay, ctc_kwargs, last_pause)
 
             # Show the text.
             if multiple:
@@ -589,7 +609,13 @@ def display_say(
                     if ctc_position == "nestled":
                         what_text.set_ctc(what_ctc)
                     elif ctc_position == "nestled-close":
-                        what_text.set_ctc([ u"\ufeff", what_ctc])
+                        what_text.set_ctc([ u"\ufeff", what_ctc, ])
+
+                if (not last_pause) and ctc:
+                    if ctc_position == "nestled":
+                        what_text.set_last_ctc(ctc)
+                    elif ctc_position == "nestled-close":
+                        what_text.set_last_ctc([ u"\ufeff", ctc, ])
 
                 if what_text.text[0] == what_string:
 
@@ -710,6 +736,8 @@ class ADVCharacter(object):
     voice_tag = None
     properties = { }
 
+    _statement_name = None
+
     # When adding a new argument here, remember to add it to copy below.
     def __init__(
             self,
@@ -777,6 +805,8 @@ class ADVCharacter(object):
             advance=d('advance'),
             )
 
+        self._statement_name = properties.pop("statement_name", None)
+
         self.properties = collections.defaultdict(dict)
 
         if kind:
@@ -1099,6 +1129,31 @@ class ADVCharacter(object):
 
         return True
 
+    def prefix_suffix(self, thing, prefix, body, suffix):
+
+        def sub(s, scope=None, force=False, translate=True):
+            return renpy.substitutions.substitute(s, scope=scope, force=force, translate=translate)[0]
+
+        thingvar_quoted = "[[" + thing + "]"
+        thingvar = "[" + thing + "]"
+
+        if not renpy.config.new_substitutions:
+            return prefix + body + suffix
+
+        # Used before Ren'Py 7.4.
+        elif renpy.config.who_what_sub_compat == 0:
+            pattern = sub(prefix + thingvar_quoted + suffix)
+            return pattern.replace(thingvar, sub(body))
+
+        # Used from Ren'Py 7.4 to Ren'Py 7.4.4
+        elif renpy.config.who_what_sub_compat == 1:
+            pattern = sub(sub(prefix) + thingvar_quoted + sub(suffix))
+            return pattern.replace(thingvar, sub(body))
+
+        # 7.4.5 on.
+        else:
+            return (sub(prefix) + sub(body) + sub(suffix))
+
     def __call__(self, what, interact=True, _call_done=True, multiple=None, **kwargs):
 
         _mode = kwargs.pop("_mode", None)
@@ -1175,21 +1230,10 @@ class ADVCharacter(object):
                 else:
                     who = renpy.python.py_eval(who)
 
-            def sub(s, scope=None, force=False, translate=True):
-                return renpy.substitutions.substitute(s, scope=scope, force=force, translate=translate)[0]
-
             if who is not None:
-                if renpy.config.new_substitutions:
-                    who_pattern = sub(sub(self.who_prefix) + "[[who]" + sub(self.who_suffix))
-                    who = who_pattern.replace("[who]", sub(who))
-                else:
-                    who = self.who_prefix + who + self.who_suffix
+                who = self.prefix_suffix("who", self.who_prefix, who, self.who_suffix)
 
-            if renpy.config.new_substitutions:
-                what_pattern = sub(sub(self.what_prefix) + "[[what]" + sub(self.what_suffix))
-                what = what_pattern.replace("[what]", sub(what, translate=True))
-            else:
-                what = self.what_prefix + what + self.what_suffix
+            what = self.prefix_suffix("what", self.what_prefix, what, self.what_suffix)
 
             # Run the add_function, to add this character to the
             # things like NVL-mode.
@@ -1232,9 +1276,10 @@ class ADVCharacter(object):
                     after = images.get_attributes(None, self.image_tag)
                     self.handle_say_transition('restore', before, after)
 
-    @property
     def statement_name(self):
-        if not (self.condition is None or renpy.python.py_eval(self.condition)):
+        if self._statement_name is not None:
+            return self._statement_name
+        elif not (self.condition is None or renpy.python.py_eval(self.condition)):
             return "say-condition-false"
         else:
             return "say"
@@ -1342,7 +1387,7 @@ class ADVCharacter(object):
 def Character(name=NotSet, kind=None, **properties):
     """
     :doc: character
-    :args: (name, kind=adv, **args)
+    :args: (name=..., kind=adv, **args)
     :name: Character
 
     Creates and returns a Character object, which controls the look
@@ -1350,8 +1395,9 @@ def Character(name=NotSet, kind=None, **properties):
 
     `name`
         If a string, the name of the character for dialogue. When
-        ``name`` is None, display of the name is omitted, as for
-        narration.
+        `name` is None, display of the name is omitted, as for
+        narration. If no name is given, the name is taken from
+        `kind`, and otherwise defaults to None.
 
     `kind`
         The Character to base this Character off of. When used, the
