@@ -1,4 +1,4 @@
-# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -21,8 +21,12 @@
 
 # This file contains code to manage focus on the display.
 
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+
 import pygame_sdl2 as pygame
 import renpy.display
+import operator
 
 
 class Focus(object):
@@ -79,11 +83,11 @@ focus_type = "mouse"
 # the focus to change.
 pending_focus_type = "mouse"
 
-
 # The current tooltip and tooltip screen.
 tooltip = None
 
-# Sets the currently focused widget.
+# Overrides the currently focused displayable.
+override = None
 
 
 def set_focused(widget, arg, screen):
@@ -104,7 +108,7 @@ def set_focused(widget, arg, screen):
     if widget is None:
         new_tooltip = None
     else:
-        new_tooltip = widget._tooltip
+        new_tooltip = widget._get_tooltip()
 
     if tooltip != new_tooltip:
         tooltip = new_tooltip
@@ -177,10 +181,20 @@ def take_focuses():
     global default_focus
     default_focus = None
 
+    global grab
+
+    grab_found = False
+
     for f in focus_list:
         if f.x is None:
             default_focus = f
 
+        if f.widget is grab:
+            grab_found = True
+
+    if not grab_found:
+        grab = None
+
     if (default_focus is not None) and (get_focused() is None):
         change_focus(default_focus, True)
 
@@ -213,6 +227,7 @@ def before_interact(roots):
     displayables.
     """
 
+    global override
     global new_grab
     global grab
 
@@ -223,7 +238,10 @@ def before_interact(roots):
         fwn.append((f, n, renpy.display.screen._current_screen))
 
     for root in roots:
-        root.find_focusable(callback, None)
+        try:
+            root.find_focusable(callback, None)
+        except renpy.display.layout.IgnoreLayers:
+            pass
 
     # Assign a full name to each focusable.
 
@@ -249,6 +267,9 @@ def before_interact(roots):
 
     fwn = fwn2
 
+    # Is this a default change?
+    default = True
+
     # We assume id(None) is not in replaced_by.
     replaced_by.pop(None, None)
 
@@ -257,6 +278,19 @@ def before_interact(roots):
 
     current = get_focused()
     current = replaced_by.get(id(current), current)
+    old_current = current
+
+    # Update the grab.
+    grab = replaced_by.get(id(grab), None)
+
+    if override is not None:
+        d = renpy.exports.get_displayable(base=True, *override)
+
+        if (d is not None) and (current is not d) and not grab:
+            current = d
+            default = False
+
+    override = None
 
     if current is not None:
         current_name = current.full_focus_name
@@ -269,16 +303,29 @@ def before_interact(roots):
         else:
             current = None
 
-    # Otherwise, focus the default widget, or nothing.
-    if current is None:
+    if grab is not None:
+        current = grab
+
+    # Otherwise, focus the default widget.
+    if (current is None) and renpy.display.interface.start_interact:
+
+        defaults = [ ]
 
         for f, n, screen in fwn:
             if f.default:
-                current = f
-                set_focused(f, None, screen)
-                break
-        else:
-            set_focused(None, None, None)
+                defaults.append((f.default, f, screen))
+
+        if defaults:
+            if len(defaults) > 1:
+                defaults.sort(key=operator.itemgetter(0))
+
+            _, f, screen = defaults[-1]
+
+            current = f
+            set_focused(f, None, screen)
+
+    if current is None:
+        set_focused(None, None, None)
 
     # Finally, mark the current widget as the focused widget, and
     # all other widgets as unfocused.
@@ -286,28 +333,30 @@ def before_interact(roots):
         if f is not current:
             renpy.display.screen.push_current_screen(screen)
             try:
-                f.unfocus(default=True)
+                if (f is old_current) and renpy.config.always_unfocus:
+                    f.unfocus(default=False)
+                else:
+                    f.unfocus(default=default)
+
             finally:
                 renpy.display.screen.pop_current_screen()
 
     if current:
         renpy.display.screen.push_current_screen(screen_of_focused)
         try:
-            current.focus(default=True)
+            current.focus(default=default)
         finally:
             renpy.display.screen.pop_current_screen()
 
-    # Update the grab.
-    grab = replaced_by.get(id(grab), None)
-
     # Clear replaced_by.
     replaced_by.clear()
 
-# This changes the focus to be the widget contained inside the new
-# focus object.
-
 
 def change_focus(newfocus, default=False):
+    """
+    Change the focus to the displayable in ``newfocus``.
+    """
+
     rv = None
 
     if grab:
@@ -371,13 +420,19 @@ def mouse_handler(ev, x, y, default=False):
 
     global pending_focus_type
 
+    if grab:
+        return
+
     if ev is not None:
         if ev.type not in (pygame.MOUSEMOTION, pygame.MOUSEBUTTONUP, pygame.MOUSEBUTTONDOWN):
             return
         else:
             pending_focus_type = "mouse"
 
-    new_focus = renpy.display.render.focus_at_point(x, y)
+    try:
+        new_focus = renpy.display.render.focus_at_point(x, y)
+    except renpy.display.layout.IgnoreLayers:
+        new_focus = None
 
     if new_focus is None:
         new_focus = default_focus
@@ -392,10 +447,13 @@ def mouse_handler(ev, x, y, default=False):
 def focus_extreme(xmul, ymul, wmul, hmul):
 
     max_focus = None
-    max_score = -(65536**2)
+    max_score = -(65536 ** 2)
 
     for f in focus_list:
 
+        if not f.widget.style.keyboard_focus:
+            continue
+
         if f.x is None:
             continue
 
@@ -415,8 +473,8 @@ def focus_extreme(xmul, ymul, wmul, hmul):
 # This calculates the distance between two points, applying
 # the given fudge factors. The distance is left squared.
 def points_dist(x0, y0, x1, y1, xfudge, yfudge):
-    return (( x0 - x1 ) * xfudge ) ** 2 + \
-           (( y0 - y1 ) * yfudge ) ** 2
+    return ((x0 - x1) * xfudge) ** 2 + \
+           ((y0 - y1) * yfudge) ** 2
 
 
 # This computes the distance between two horizontal lines. (So the
@@ -489,7 +547,15 @@ def focus_nearest(from_x0, from_y0, from_x1, from_y1,
     current = get_focused()
 
     if not current:
-        change_focus(focus_list[0])
+
+        for f in focus_list:
+
+            if not f.widget.style.keyboard_focus:
+                continue
+
+            change_focus(f)
+            return
+
         return
 
     # Find the current focus.
