@@ -1,4 +1,4 @@
-# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -22,14 +22,17 @@
 # This file contains code responsible for managing the execution of a
 # renpy object, as well as the context object.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
+from future.utils import reraise
+
 import sys
 import time
 
 import renpy.display
 import renpy.test
 
-pyast = __import__("ast", { })
+import ast as pyast
 
 # The number of statements that have been run since the last infinite loop
 # check.
@@ -106,6 +109,9 @@ class LineLogEntry(object):
 
         return (self.filename == other.filename) and (self.line == other.line) and (self.node is other.node)
 
+    def __ne__(self, other):
+        return not (self == other)
+
 
 class Context(renpy.object.Object):
     """
@@ -130,7 +136,7 @@ class Context(renpy.object.Object):
     does participates in rollback.
     """
 
-    __version__ = 14
+    __version__ = 16
 
     nosave = [ 'next_node' ]
 
@@ -138,6 +144,26 @@ class Context(renpy.object.Object):
 
     force_checkpoint = False
 
+    come_from_name = None
+    come_from_label = None
+
+    temporary_attributes = None
+
+    deferred_translate_identifier = None
+
+    def __repr__(self):
+
+        if not self.current:
+            return "<Context>"
+
+        node = renpy.game.script.lookup(self.current)
+
+        return "<Context: {}:{} {!r}>".format(
+            node.filename,
+            node.linenumber,
+            node.diff_info(),
+            )
+
     def after_upgrade(self, version):
         if version < 1:
             self.scene_lists.image_predict_info = self.predict_info.images
@@ -176,15 +202,18 @@ class Context(renpy.object.Object):
         if version < 11:
             self.say_attributes = None
 
-        if version < 12:
-            self.translate_block_language = None
-
         if version < 13:
             self.line_log = [ ]
 
         if version < 14:
             self.movie = { }
 
+        if version < 15:
+            self.abnormal_stack = [ False ] * len(self.return_stack)
+
+        if version < 16:
+            self.alternate_translate_identifier = None
+
     def __init__(self, rollback, context=None, clear=False):
         """
         `clear`
@@ -197,6 +226,9 @@ class Context(renpy.object.Object):
         self.call_location_stack = [ ]
         self.return_stack = [ ]
 
+        # The value of abnormal at the time of the call.
+        self.abnormal_stack = [ ]
+
         # Two deeper then the return stack and call location stack.
         # 1 deeper is for the context top-level, 2 deeper is for
         # _args, _kwargs, and _return.
@@ -220,7 +252,7 @@ class Context(renpy.object.Object):
 
         # A map from the name of a music channel to the MusicContext
         # object corresponding to that channel.
-        self.music = renpy.python.RevertableDict()
+        self.music = { }
 
         # True if we're in the middle of a call to ui.interact. This
         # will cause Ren'Py to generate an error if we call ui.interact
@@ -239,6 +271,7 @@ class Context(renpy.object.Object):
 
         # The attributes that are used by the current say statement.
         self.say_attributes = None
+        self.temporary_attributes = None
 
         # A list of lines that were run since the last time this log was
         # cleared.
@@ -248,7 +281,7 @@ class Context(renpy.object.Object):
         # executed?
         self.force_checkpoint = False
 
-        # A mapt from a channel to the Movie playing on that channel.
+        # A map from a channel to the Movie playing on that channel.
         self.movie = { }
 
         if context:
@@ -257,9 +290,7 @@ class Context(renpy.object.Object):
 
             vars(self.info).update(vars(context.info))
 
-            for k, v in context.music.iteritems():
-                self.music[k] = v.copy()
-
+            self.music = dict(context.music)
             self.movie = dict(context.movie)
 
             self.images = renpy.display.image.ShownImageInfo(context.images)
@@ -270,6 +301,9 @@ class Context(renpy.object.Object):
 
         self.scene_lists = renpy.display.core.SceneLists(oldsl, self.images)
 
+        for i in renpy.config.context_copy_remove_screens:
+            self.scene_lists.remove("screens", i, None)
+
         self.make_dynamic([ "_return", "_args", "_kwargs", "mouse_visible", "suppress_overlay", "_side_image_attributes" ])
         self.dynamic_stack.append({ })
 
@@ -282,13 +316,29 @@ class Context(renpy.object.Object):
         self.use_modes = True
 
         # The language we started with.
-        self.translate_language = None
+        self.translate_language = renpy.game.preferences.language
 
         # The identifier of the current translate block.
         self.translate_identifier = None
 
-        # The language of the current translate block.
-        self.translate_block_language = None
+        # The alternate identifier of the current translate block.
+        self.alternate_translate_identifier = None
+
+        # The translate identifier of the last say statement with
+        # interact = False.
+        self.deferred_translate_identifier = None
+
+    def replace_node(self, old, new):
+
+        def replace_one(name):
+            n = renpy.game.script.lookup(name)
+            if n is old:
+                return new.name
+
+            return name
+
+        self.current = replace_one(self.current)
+        self.return_stack = [ replace_one(i) for i in self.return_stack ]
 
     def make_dynamic(self, names, context=False):
         """
@@ -326,9 +376,9 @@ class Context(renpy.object.Object):
 
         dynamic = self.dynamic_stack.pop()
 
-        for k, v in dynamic.iteritems():
+        for k, v in dynamic.items():
             if isinstance(v, Delete):
-                del store[k]
+                store.pop(k, None)
             else:
                 store[k] = v
 
@@ -345,7 +395,7 @@ class Context(renpy.object.Object):
 
         for dynamic in reversed(self.dynamic_stack):
 
-            for k, v in dynamic.iteritems():
+            for k, v in dynamic.items():
                 name = "store." + k
 
                 if isinstance(v, Delete) and (name in roots):
@@ -414,6 +464,14 @@ class Context(renpy.object.Object):
         code = compile(module, node.filename, 'exec')
         exec(code)
 
+    def come_from(self, name, label):
+        """
+        When control reaches name, call label. Only for internal use.
+        """
+
+        self.come_from_name = name
+        self.come_from_label = label
+
     def run(self, node=None):
         """
         Executes as many nodes as possible in the current context. If the
@@ -431,8 +489,22 @@ class Context(renpy.object.Object):
         developer = renpy.config.developer
         tracing = sys.gettrace() is not None
 
+        # Is this the first time through the loop?
+        first = True
+
         while node:
 
+            if node.name == self.come_from_name:
+                self.come_from_name = None
+                node = self.call(self.come_from_label, return_site=node.name)
+                self.make_dynamic([ "_return", "_begin_rollback" ])
+                renpy.store._begin_rollback = False
+
+            this_node = node
+            type_node_name = type(node).__name__
+
+            renpy.plog(1, "--- start {} ({}:{})", type_node_name, node.filename, node.linenumber)
+
             self.current = node.name
             self.last_abnormal = self.abnormal
             self.abnormal = False
@@ -444,12 +516,34 @@ class Context(renpy.object.Object):
                 if ll_entry not in self.line_log:
                     self.line_log.append(ll_entry)
 
-            if self.rollback and renpy.game.log:
-                renpy.game.log.begin()
+            if not renpy.store._begin_rollback:
+                update_rollback = False
+                force_rollback = False
+            elif first or self.force_checkpoint or (node.rollback == "force"):
+                update_rollback = True
+                force_rollback = True
+            elif not renpy.config.all_nodes_rollback and (node.rollback == "never"):
+                update_rollback = False
+                force_rollback = False
+            else:
+                update_rollback = True
+                force_rollback = False
 
-            if self.force_checkpoint:
-                renpy.game.log.checkpoint(hard=False)
-                self.force_checkpoint = False
+            # Force a new rollback to start to match things in the forward log.
+            if renpy.game.log.forward and renpy.game.log.forward[0][0] == node.name:
+                update_rollback = True
+                force_rollback = True
+
+            first = False
+
+            if update_rollback:
+
+                if self.rollback and renpy.game.log:
+                    renpy.game.log.begin(force=force_rollback)
+
+                if self.rollback and self.force_checkpoint:
+                    renpy.game.log.force_checkpoint = True
+                    self.force_checkpoint = False
 
             self.seen = False
 
@@ -465,8 +559,13 @@ class Context(renpy.object.Object):
                     renpy.game.exception_info = "While running game code:"
 
                     self.next_node = None
+
+                    renpy.plog(2, "    before execute {} ({}:{})", type_node_name, node.filename, node.linenumber)
+
                     node.execute()
 
+                    renpy.plog(2, "    after execute {} ({}:{})", type_node_name, node.filename, node.linenumber)
+
                     if developer and self.next_node:
                         self.check_stacks()
 
@@ -484,14 +583,21 @@ class Context(renpy.object.Object):
                     short, full, traceback_fn = renpy.error.report_exception(e, editor=False)
 
                     try:
+                        handled = False
+
                         if self.exception_handler is not None:
                             self.exception_handler(short, full, traceback_fn)
-                        elif renpy.display.error.report_exception(short, full, traceback_fn):
-                            raise
+                            handled = True
+                        elif renpy.config.exception_handler is not None:
+                            handled = renpy.config.exception_handler(short, full, traceback_fn)
+
+                        if not handled:
+                            if renpy.display.error.report_exception(short, full, traceback_fn):
+                                raise
                     except renpy.game.CONTROL_EXCEPTIONS as ce:
                         raise ce
                     except Exception as ce:
-                        raise exc_info[0], exc_info[1], exc_info[2]
+                        reraise(exc_info[0], exc_info[1], exc_info[2])
 
                 node = self.next_node
 
@@ -502,7 +608,7 @@ class Context(renpy.object.Object):
             except renpy.game.CallException as e:
 
                 if e.from_current:
-                    return_site = node.name
+                    return_site = getattr(node, "statement_start", node).name
                 else:
                     if self.next_node is None:
                         raise Exception("renpy.call can't be used when the next node is undefined.")
@@ -514,11 +620,13 @@ class Context(renpy.object.Object):
                 renpy.store._kwargs = e.kwargs
 
             if self.seen:
-                renpy.game.persistent._seen_ever[self.current] = True  # @UndefinedVariable
+                renpy.game.persistent._seen_ever[self.current] = True # @UndefinedVariable
                 renpy.game.seen_session[self.current] = True
 
-            if self.rollback and renpy.game.log:
-                renpy.game.log.complete()
+            renpy.plog(2, "    end {} ({}:{})", type_node_name, this_node.filename, this_node.linenumber)
+
+        if self.rollback and renpy.game.log:
+            renpy.game.log.complete()
 
     def mark_seen(self):
         """
@@ -542,6 +650,7 @@ class Context(renpy.object.Object):
 
         self.return_stack.append(return_site)
         self.dynamic_stack.append({ })
+        self.abnormal_stack.append(self.last_abnormal)
         self.current = label
 
         self.make_dynamic([ "_args", "_kwargs" ])
@@ -564,6 +673,7 @@ class Context(renpy.object.Object):
         self.return_stack.pop()
         self.call_location_stack.pop()
         self.pop_dynamic()
+        self.abnormal_stack.pop()
 
     def lookup_return(self, pop=True):
         """
@@ -582,18 +692,34 @@ class Context(renpy.object.Object):
 
             if node is None:
 
-                if renpy.config.developer:
-                    raise Exception("Could not find return label {!r}.".format(self.return_stack[-1]))
+                if not pop:
+                    return None
 
-                self.return_stack.pop()
-                self.call_location_stack.pop()
-                self.pop_dynamic()
+                # If we can't find anything, try to recover.
 
-                continue
+                if renpy.config.return_not_found_label:
+
+                    while len(self.return_stack) > 1:
+                        self.pop_call()
+
+                    node = renpy.game.script.lookup(renpy.config.return_not_found_label)
+
+                else:
+
+                    if renpy.config.developer:
+                        raise Exception("Could not find return label {!r}.".format(self.return_stack[-1]))
+
+                    self.return_stack.pop()
+                    self.call_location_stack.pop()
+                    self.pop_dynamic()
+                    self.abnormal = self.abnormal_stack.pop()
+
+                    continue
 
             if pop:
                 self.return_stack.pop()
                 self.call_location_stack.pop()
+                self.abnormal = self.abnormal_stack.pop()
 
             return node
 
@@ -616,6 +742,10 @@ class Context(renpy.object.Object):
         rv.translate_language = self.translate_language
         rv.translate_identifier = self.translate_identifier
 
+        rv.abnormal = self.abnormal
+        rv.last_abnormal = self.last_abnormal
+        rv.abnormal_stack = list(self.abnormal_stack)
+
         return rv
 
     def predict_call(self, label, return_site):
@@ -675,7 +805,7 @@ class Context(renpy.object.Object):
         for label in renpy.config.predict_statements_callback(self.current):
 
             if not renpy.game.script.has_label(label):
-                return
+                continue
 
             node = renpy.game.script.lookup(label)
 
@@ -708,12 +838,12 @@ class Context(renpy.object.Object):
 
             except:
 
-                if renpy.config.debug_image_cache:
+                if renpy.config.debug_prediction:
                     import traceback
 
-                    print()
-                    traceback.print_exc()
                     print("While predicting images.")
+                    traceback.print_exc()
+                    print()
 
             self.images = old_images
             self.predict_return_stack = None
@@ -736,7 +866,7 @@ class Context(renpy.object.Object):
             return False
 
         if ever:
-            seen = renpy.game.persistent._seen_ever  # @UndefinedVariable
+            seen = renpy.game.persistent._seen_ever # @UndefinedVariable
         else:
             seen = renpy.game.seen_session
 
@@ -780,7 +910,8 @@ def run_context(top):
     the RestartContext and RestartTopContext exceptions.
     """
 
-    label = None
+    if renpy.config.context_callback is not None:
+        renpy.config.context_callback()
 
     while True:
 
@@ -788,10 +919,6 @@ def run_context(top):
 
             context = renpy.game.context()
 
-            if label and renpy.game.script.has_label(label):
-                context.call(label)
-            label = None
-
             context.run()
 
             rv = renpy.store._return
@@ -801,24 +928,10 @@ def run_context(top):
             return rv
 
         except renpy.game.RestartContext as e:
-
-            # Apply defaults.
-            renpy.exports.execute_default_statement(False)
-
-            if e.label:
-                label = e.label
-
             continue
 
         except renpy.game.RestartTopContext as e:
             if top:
-
-                # Apply defaults.
-                renpy.exports.execute_default_statement(False)
-
-                if e.label:
-                    label = e.label
-
                 continue
 
             else:
