@@ -52,8 +52,6 @@ def get_line_text(filename, lineno):
     import linecache
     full_filename = renpy.exports.unelide_filename(filename)
 
-    print(full_filename, linecache.getline(full_filename, lineno))
-
     try:
         line = linecache.getline(full_filename, lineno) or "\n"
         line = line.decode("utf-8")
@@ -1619,6 +1617,7 @@ def parse_menu(stmtl, loc, arguments):
 
     say_who = None
     say_what = None
+    say_loc = None
 
     # Tuples of (label, condition, block)
     items = [ ]
@@ -1660,6 +1659,7 @@ def parse_menu(stmtl, loc, arguments):
             has_say = True
             say_who = who
             say_what = what
+            say_loc = l.get_location()
 
             continue
 
@@ -1711,7 +1711,7 @@ def parse_menu(stmtl, loc, arguments):
 
     rv = [ ]
     if has_say:
-        rv.append(ast.Say(loc, say_who, say_what, None, interact=False))
+        rv.append(ast.Say(say_loc, say_who, say_what, None, interact=False))
 
     rv.append(ast.Menu(loc, items, set, with_, has_say or has_caption, arguments, item_arguments))
 
@@ -1783,6 +1783,8 @@ def parse_parameters(l):
             if l.match(r'='):
                 l.skip_whitespace()
                 default = l.delimited_python("),")
+                if not default.strip():
+                    l.error("Empty parameter default.")
             else:
                 default = None
 
@@ -1808,6 +1810,8 @@ def parse_arguments(l):
     extrakw = None
     extrapos = None
 
+    has_kw = False
+
     if not l.match(r'\('):
         return None
 
@@ -1838,6 +1842,11 @@ def parse_arguments(l):
                 l.revert(state)
                 name = None
 
+            if name:
+                has_kw = True
+            elif has_kw:
+                l.error("positional argument follows keyword argument")
+
             l.skip_whitespace()
             arguments.append((name, l.delimited_python("),")))
 
