@@ -1,4 +1,4 @@
-# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2021 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,7 +19,8 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-from __future__ import print_function
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import *
 
 import math
 import renpy.display
@@ -31,7 +32,7 @@ import renpy.text.texwrap as texwrap
 import renpy.text.font as font
 import renpy.text.extras as extras
 
-from _renpybidi import log2vis, WRTL, RTL, ON  # @UnresolvedImport
+from _renpybidi import log2vis, WRTL, RTL, ON # @UnresolvedImport
 
 BASELINE = -65536
 
@@ -188,12 +189,14 @@ class TextSegment(object):
             self.ruby_bottom = source.ruby_bottom
             self.hinting = source.hinting
             self.outline_color = source.outline_color
+            self.ignore = source.ignore
 
         else:
             self.hyperlink = 0
             self.cps = 0
             self.ruby_top = False
             self.ruby_bottom = False
+            self.ignore = False
 
     def __repr__(self):
         return "<TextSegment font={font}, size={size}, bold={bold}, italic={italic}, underline={underline}, color={color}, black_color={black_color}, hyperlink={hyperlink}, vertical={vertical}>".format(**self.__dict__)
@@ -239,6 +242,9 @@ class TextSegment(object):
         Return the list of glyphs corresponding to unicode string s.
         """
 
+        if self.ignore:
+            return [ ]
+
         fo = font.get_font(self.font, self.size, self.bold, self.italic, 0, self.antialias, self.vertical, self.hinting, layout.oversample)
         rv = fo.glyphs(s)
 
@@ -293,13 +299,28 @@ class TextSegment(object):
         on the font group.
         """
 
-        if not isinstance(self.font, font.FontGroup):
-            yield (self, s)
+        tf = self.font
+
+        font_transform = renpy.game.preferences.font_transform
+        if font_transform is not None:
+            font_func = renpy.config.font_transforms.get(font_transform, None)
+            if font_func is not None:
+                tf = font_func(tf)
+
+        if not isinstance(tf, font.FontGroup):
+
+            if self.font is tf:
+                yield (self , s)
+            else:
+                seg = TextSegment(self)
+                seg.font = tf
+                yield (seg, s)
+
             return
 
         segs = { }
 
-        for f, ss in self.font.segment(s):
+        for f, ss in tf.segment(s):
 
             seg = segs.get(f, None)
 
@@ -370,7 +391,7 @@ class SpaceSegment(object):
 
 class DisplayableSegment(object):
     """
-    A segment that's used to render horizontal or vertical whitespace.
+    A segment that's used to render displayables.
     """
 
     def __init__(self, ts, d, renders):
@@ -384,6 +405,9 @@ class DisplayableSegment(object):
 
         self.width, self.height = rend.get_size()
 
+        if isinstance(d, renpy.display.behavior.CaretBlink):
+            self.width = 0
+
         self.hyperlink = ts.hyperlink
         self.cps = ts.cps
         self.ruby_top = ts.ruby_top
@@ -420,7 +444,6 @@ class DisplayableSegment(object):
         glyph = glyphs[0]
 
         if di.displayable_blits is not None:
-
             di.displayable_blits.append((self.d, glyph.x, glyph.y, glyph.width, glyph.ascent, glyph.line_spacing, glyph.time))
 
     def assign_times(self, gt, glyphs):
@@ -500,6 +523,8 @@ class Layout(object):
 
             self.outline_step = text.style.outline_scaling != "linear"
 
+            self.pixel_perfect = True
+
         else:
 
             self.oversample = 1.0
@@ -507,6 +532,8 @@ class Layout(object):
             self.forward = renpy.display.render.IDENTITY
             self.outline_step = True
 
+            self.pixel_perfect = False
+
         style = text.style
 
         self.line_overlap_split = self.scale_int(style.line_overlap_split)
@@ -528,12 +555,13 @@ class Layout(object):
         # data from one Layout to another.
         self.paragraph_glyphs = [ ]
 
-        width = self.scale_int(width)
-        height = self.scale_int(height)
-
+        # The virtual width and height offered to this Layout.
         self.width = width
         self.height = height
 
+        width = self.scale_int(width)
+        height = self.scale_int(height)
+
         # Figure out outlines and other info.
         outlines, xborder, yborder, xoffset, yoffset = self.figure_outlines(style)
         self.outlines = outlines
@@ -626,7 +654,7 @@ class Layout(object):
             self.paragraph_glyphs.append(list(par_glyphs))
 
             if splits_from:
-                textsupport.copy_splits(splits_from.paragraph_glyphs[p_num], par_glyphs)  # @UndefinedVariable
+                textsupport.copy_splits(splits_from.paragraph_glyphs[p_num], par_glyphs) # @UndefinedVariable
 
             else:
 
@@ -733,7 +761,7 @@ class Layout(object):
             elif adjust_spacing == "vertical":
                 target_x_delta = 0.0
 
-            textsupport.tweak_glyph_spacing(all_glyphs, lines, target_x_delta, target_y_delta, maxx, y)  # @UndefinedVariable
+            textsupport.tweak_glyph_spacing(all_glyphs, lines, target_x_delta, target_y_delta, maxx, y) # @UndefinedVariable
 
             maxx = target_x
             y = target_y
@@ -783,27 +811,57 @@ class Layout(object):
             if key in self.textures:
                 continue
 
-            # Create the texture.
-            surf = renpy.display.pgrender.surface((sw + o, sh + o), True)
-
-            di.surface = surf
-            di.override_color = color
-            di.outline = o
-
             if color == None:
                 self.displayable_blits = [ ]
                 di.displayable_blits = self.displayable_blits
             else:
                 di.displayable_blits = None
 
+            # Create the texture.
+
+            tw = int(sw + o)
+            th = int(sh + o)
+
+            # If not a multiple of 32, round up.
+            tw = (tw | 0x1f) + 1 if (tw & 0x1f) else tw
+            th = (th | 0x1f) + 1 if (th & 0x1f) else th
+
+            surf = renpy.display.pgrender.surface((tw, th), True)
+
+            if renpy.game.preferences.high_contrast:
+                if color:
+                    surf.fill(color)
+                else:
+                    prefix = style.prefix
+                    prefix_color = style.color
+                    style.set_prefix("idle_")
+                    idle_color = style.color
+                    style.set_prefix(prefix)
+
+                    color = (255, 255, 255, 255)
+
+                    if idle_color != prefix_color:
+
+                        if "hover" in prefix:
+                            color = (255, 255, 224, 255)
+                        elif "selected" in style.prefix:
+                            color = (224, 255, 255, 255)
+
+            di.surface = surf
+            di.override_color = color
+            di.outline = o
+
             for ts, glyphs in par_seg_glyphs:
                 if ts is self.end_segment:
                     break
 
                 ts.draw(glyphs, di, self.add_left, self.add_top, self)
 
+            if renpy.config.debug_text_alignment:
+                self.make_alignment_grid(surf)
+
             renpy.display.draw.mutated_surface(surf)
-            tex = renpy.display.draw.load_texture(surf)
+            tex = renpy.display.draw.load_texture(surf, properties={ "mipmap" : renpy.config.mipmap_text if (style.mipmap is None) else style.mipmap })
 
             self.textures[key] = tex
 
@@ -832,6 +890,22 @@ class Layout(object):
                 renpy.display.to_log.write("     Available: (%d, %d) Laid-out: (%d, %d)", width, height, sw, sh)
                 renpy.display.to_log.write("     Text: %r", text.text)
 
+    def make_alignment_grid(self, surf):
+        w, h = surf.get_size()
+
+        for x in range(w):
+            for y in range(h):
+
+                if surf.get_at((x, y))[3] > 0:
+                    continue
+
+                if x == 0 or y == 0 or x == w - 1 or y == h - 1:
+                    surf.set_at((x, y), (255, 0, 0, 255))
+                elif (x ^ y) & 1:
+                    surf.set_at((x, y), (255, 255, 255, 255))
+                else:
+                    surf.set_at((x, y), (0, 0, 0, 255))
+
     def scale(self, n):
         if n is None:
             return n
@@ -859,13 +933,13 @@ class Layout(object):
             if self.oversample < 1:
                 return n
 
-            return n * int(self.oversample)
+            return int(n * int(self.oversample))
 
         else:
             if n == 0:
                 return 0
 
-            rv = round(n * self.oversample)
+            rv = int(round(n * self.oversample))
 
             if n < 0 and rv > -1:
                 rv = -1
@@ -920,7 +994,7 @@ class Layout(object):
 
             line.extend(tss[-1].subsegment(u"\u200B"))
 
-        for type, text in tokens:  # @ReservedAssignment
+        for type, text in tokens: # @ReservedAssignment
 
             try:
 
@@ -936,6 +1010,11 @@ class Layout(object):
                     continue
 
                 elif type == TEXT:
+
+                    if (text_displayable.mask is not None):
+                        if text != u"\u200b":
+                            text = text_displayable.mask * len(text)
+
                     line.extend(tss[-1].subsegment(text))
                     continue
 
@@ -951,6 +1030,12 @@ class Layout(object):
                     tss.pop()
 
                     if not tss:
+
+                        tag = tag[1:]
+
+                        if not renpy.text.extras.text_tags.get(tag, True):
+                            raise Exception("{/%s} isn't valid, since the %s text tag doesn't take a close tag." % (tag, tag))
+
                         raise Exception("%r closes a text tag that isn't open." % text)
 
                 elif tag == "_start":
@@ -1001,6 +1086,9 @@ class Layout(object):
                 elif tag == "fast":
                     pass
 
+                elif tag == "done":
+                    pass
+
                 elif tag == "nw":
                     pass
 
@@ -1157,6 +1245,13 @@ class Layout(object):
                     ts = push()
                     ts.vertical = False
 
+                elif tag == "alt":
+                    ts = push()
+                    ts.ignore = True
+
+                elif tag == "noalt":
+                    ts = push()
+
                 elif tag[0] == "#":
                     pass
 
@@ -1187,7 +1282,7 @@ class Layout(object):
         l = [ ]
 
         for ts, s in p:
-            s, direction = log2vis(s, direction)
+            s, direction = log2vis(unicode(s), direction)
             l.append((ts, s))
 
         rtl = (direction == RTL or direction == WRTL)
@@ -1206,7 +1301,7 @@ class Layout(object):
         style_outlines = style.outlines
         dslist = style.drop_shadow
 
-        if not style_outlines and not dslist:
+        if (not style_outlines) and (not dslist) and not renpy.game.preferences.high_contrast:
             return [ (0, None, 0, 0) ], 0, 0, 0, 0
 
         outlines = [ ]
@@ -1248,6 +1343,9 @@ class Layout(object):
 
         outlines.append((0, None, 0, 0))
 
+        if renpy.game.preferences.high_contrast:
+            outlines = [ (2, (0, 0, 0, 255), 0, 0), (0, None, 0, 0) ]
+
         return outlines, right - left, bottom - top, -left, -top
 
     def blits_typewriter(self, st):
@@ -1311,13 +1409,13 @@ class Layout(object):
             if g.x + g.advance > max_x:
                 max_x = g.x + g.advance
 
-            if g.x  < min_x:
+            if g.x < min_x:
                 min_x = g.x
 
         ly = min(l.y + l.height + self.line_overlap_split, max_height)
 
         if min_x < max_x:
-            rv.append(Blit(min_x, max_y, max_x - min_x, ly - max_y, left=left, right=right, top=top, bottom=(l is self.lines[-1]) ))
+            rv.append(Blit(min_x, max_y, max_x - min_x, ly - max_y, left=left, right=right, top=top, bottom=(l is self.lines[-1])))
 
         return rv
 
@@ -1395,7 +1493,7 @@ class Text(renpy.display.core.Displayable):
     """
     :name: Text
     :doc: text
-    :args: (text, slow=None, scope=None, substitute=None, slow_done=None, **properties)
+    :args: (text, slow=None, scope=None, substitute=None, slow_done=None, mipmap=None, **properties)
 
     A displayable that displays text on the screen.
 
@@ -1424,6 +1522,8 @@ class Text(renpy.display.core.Displayable):
     locked = False
 
     language = None
+    mask = None
+    last_ctc = None
 
     def after_upgrade(self, version):
 
@@ -1441,7 +1541,7 @@ class Text(renpy.display.core.Displayable):
             self.end = None
             self.dirty = True
 
-    def __init__(self, text, slow=None, scope=None, substitute=None, slow_done=None, replaces=None, **properties):
+    def __init__(self, text, slow=None, scope=None, substitute=None, slow_done=None, replaces=None, mask=None, **properties):
 
         super(Text, self).__init__(**properties)
 
@@ -1467,6 +1567,9 @@ class Text(renpy.display.core.Displayable):
         # The text, after substitutions.
         self.text = None
 
+        # A mask, for passwords and such.
+        self.mask = mask
+
         # Sets the text we're showing, and performs substitutions.
         self.set_text(text, scope, substitute)
 
@@ -1487,7 +1590,7 @@ class Text(renpy.display.core.Displayable):
         self.start = None
         self.end = None
 
-        if replaces is not None:
+        if isinstance(replaces, Text):
             self.slow = replaces.slow
             self.slow_done = replaces.slow_done
             self.ctc = replaces.ctc
@@ -1590,10 +1693,8 @@ class Text(renpy.display.core.Displayable):
                     i, did_sub = renpy.substitutions.substitute(i, scope, substitute)
                     uses_scope = uses_scope or did_sub
 
-                if isinstance(i, str):
-                    i = unicode(i, "utf-8", "replace")
-                else:
-                    i = unicode(i)
+                if isinstance(i, bytes):
+                    i = str(i, "utf-8", "replace")
 
             new_text.append(i)
 
@@ -1625,6 +1726,10 @@ class Text(renpy.display.core.Displayable):
         self.ctc = ctc
         self.dirty = True
 
+    def set_last_ctc(self, last_ctc):
+        self.last_ctc = last_ctc
+        self.dirty = True
+
     def update(self):
         """
         This needs to be called after text has been updated, but before
@@ -1673,12 +1778,21 @@ class Text(renpy.display.core.Displayable):
         else:
             # Add the CTC.
             if self.ctc is not None:
-                text.append(self.ctc)
+                if isinstance(self.ctc, list):
+                    text.extend(self.ctc)
+                else:
+                    text.append(self.ctc)
+
+        if self.last_ctc is not None:
+            if isinstance(self.last_ctc, list):
+                text.extend(self.last_ctc)
+            else:
+                text.append(self.last_ctc)
 
         # Tokenize the text.
         tokens = self.tokenize(text)
 
-        if renpy.config.custom_text_tags or (renpy.config.replace_text is not None):
+        if renpy.config.custom_text_tags or renpy.config.self_closing_custom_text_tags or (renpy.config.replace_text is not None):
             tokens = self.apply_custom_tags(tokens)
 
         # self.tokens is a list of pairs, where the first component of
@@ -1715,9 +1829,10 @@ class Text(renpy.display.core.Displayable):
             rv.append(i)
 
         rv = "".join(rv)
+        rv, _, _ = rv.partition("{done}")
         _, _, rv = rv.rpartition("{fast}")
 
-        rv = renpy.translation.dialogue.notags_filter(rv)
+        rv = renpy.text.extras.filter_alt_text(rv)
 
         alt = self.style.alt
 
@@ -2134,6 +2249,9 @@ class Text(renpy.display.core.Displayable):
             vrv.blit(rv, (rv.height, 0))
             rv = vrv
 
+        if layout.pixel_perfect:
+            rv.properties = { "pixel_perfect" : True }
+
         return rv
 
     def tokenize(self, text):
@@ -2145,11 +2263,11 @@ class Text(renpy.display.core.Displayable):
 
         for i in text:
 
-            if isinstance(i, unicode):
+            if isinstance(i, str):
                 tokens.extend(textsupport.tokenize(i))
 
-            elif isinstance(i, str):
-                tokens.extend(textsupport.tokenize(unicode(i)))
+            elif isinstance(i, basestring):
+                tokens.extend(textsupport.tokenize(str(i)))
 
             elif isinstance(i, renpy.display.core.Displayable):
                 tokens.append((DISPLAYABLE, i))
@@ -2172,7 +2290,7 @@ class Text(renpy.display.core.Displayable):
             kind, text = t
 
             if kind == TEXT and renpy.config.replace_text:
-                rv.append((TEXT, unicode(renpy.config.replace_text(text))))
+                rv.append((TEXT, str(renpy.config.replace_text(text))))
 
             elif kind != TAG:
                 rv.append(t)
@@ -2183,48 +2301,60 @@ class Text(renpy.display.core.Displayable):
 
                 func = renpy.config.custom_text_tags.get(tag, None)
 
+                if func is None:
+                    func = renpy.config.self_closing_custom_text_tags.get(tag, None)
+                    self_closing = True
+                else:
+                    self_closing = False
+
                 if func is None:
                     rv.append(t)
                     continue
 
-                # The contents of this tag.
-                contents = [ ]
+                if not self_closing:
 
-                # The close tag we're lookin for.
-                close = "/" + tag
+                    # The contents of this tag.
+                    contents = [ ]
 
-                # The number of open tags.
-                count = 1
+                    # The close tag we're lookin for.
+                    close = "/" + tag
 
-                while tokens:
+                    # The number of open tags.
+                    count = 1
 
-                    # Count the number of `tag` tags that are still open.
-                    t2 = tokens.pop(0)
+                    while tokens:
 
-                    kind2, text2 = t2
+                        # Count the number of `tag` tags that are still open.
+                        t2 = tokens.pop(0)
 
-                    if kind2 == TAG:
-                        tag2, _, _ = text2.partition("=")
+                        kind2, text2 = t2
 
-                        if tag2 == tag:
-                            count += 1
-                        elif tag2 == close:
-                            count -= 1
-                            if not count:
-                                break
+                        if kind2 == TAG:
+                            tag2, _, _ = text2.partition("=")
 
-                    contents.append(t2)
+                            if tag2 == tag:
+                                count += 1
+                            elif tag2 == close:
+                                count -= 1
+                                if not count:
+                                    break
 
-                if count:
-                    raise Exception("Text ended while the '{}' text tag was still open.".format(tag))
+                        contents.append(t2)
 
-                new_contents = func(tag, value, contents)
+                    if count:
+                        raise Exception("Text ended while the '{}' text tag was still open.".format(tag))
+
+                    new_contents = func(tag, value, contents)
+
+                else:
+
+                    new_contents = func(tag, value)
 
                 new_tokens = [ ]
 
                 for kind2, text2 in new_contents:
-                    if isinstance(text2, str):
-                        text2 = unicode(text2)
+                    if isinstance(text2, bytes):
+                        text2 = str(text2)
 
                     new_tokens.append((kind2, text2))
 
